
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include <windows.h>


//	maximum block size
#define K 4096
//	iterations per algorithm per block size
#define N 200

extern "C" {
	bool copytest_a(void * dst, const void * src, size_t size);
	LONGLONG rdtsc();
	void copytest_a_getinfo(const char ** name, size_t * blocksize);
}

int g;

static void
clearcache()
{
	long * cc = (long *)malloc(800000);
  int x = g;
	for( int ix = 0; ix < 50000; ++ix ) {
    x += cc[ ix ] + cc[ ix+50000 ] + cc[ ix+100000 ] + cc[ ix+150000 ];
	}
	free(cc);
  g = x;
}

int
main(
	int argc,
	char * argv[])
{
	double cpuFreq = 0.;

//	assert( !copytest_a(0, 0, 1) );
	assert( copytest_a(0, 0, 0) );
	assert( K >= 16 );

	char * src = (char *)malloc(1024*K+64);
	char * dst = (char *)malloc(1024*K+64);

	memset(src, 2, 1024*K+64);
	memset(dst, 1, 1024*K+64);

	src += 32-((long)src & 31);
	dst += 32-((long)dst & 31);

	copytest_a(dst, src, 16384);

	assert( dst[-1] == 1 );
	assert( dst[0] == 2 );
	assert( dst[16383] == 2 );
	assert( dst[16384] == 1 );

	{
		const char * name;
		size_t blocksize;
		copytest_a_getinfo(&name, &blocksize);
		printf( "comparing %s (blocksize %ld) to memcpy\n", name, blocksize );
	}
	{
		LARGE_INTEGER pc1, pc2, pcFreq;
		QueryPerformanceFrequency( &pcFreq );
		QueryPerformanceCounter( &pc1 );
		LONGLONG uno = rdtsc();
		{for( int k = 0; k < 50; ++k ) {
			clearcache();
		}}
		QueryPerformanceCounter( &pc2 );
		LONGLONG dos = rdtsc();
		double freqCalc = double( dos-uno );
		double factor = double( pc2.QuadPart-pc1.QuadPart );
		factor /= pcFreq.QuadPart;
		freqCalc /= factor;
		printf( "calculated CPU speed: %.1f MHz\n", freqCalc/1000000. );
    cpuFreq = freqCalc;
	}

	printf( "size\tasm\tthru-a\tmemcpy\tthru-m\n" );

	{for( int k = 1; k <= K; k <<= 1 ) {
		LONGLONG best_a = 10000000000;
		{for( int ix = 0; ix < N; ++ix ) {
			clearcache();
			LONGLONG stop, start = rdtsc();
			copytest_a(dst, src, 1024*k);
			stop = rdtsc() - start;
			if( stop < best_a ) best_a = stop;
		}}
		LONGLONG best_b = 10000000000;
		{for( int ix = 0; ix < N; ++ix ) {
			clearcache();
			LONGLONG stop, start = rdtsc();
			memcpy(dst, src, 1024*k);
			stop = rdtsc() - start;
			if( stop < best_b ) best_b = stop;
		}}
		double mbps_a = (double)k/1024.;
		mbps_a *= cpuFreq;
		mbps_a /= (double)best_a;
		double mbps_b = (double)k/1024.;
		mbps_b *= cpuFreq;
		mbps_b /= (double)best_b;
		printf("%d kB\t%d c\t%.2f MB/s\t%d c\t%.2f MB/s\n",
			k,
			(int)best_a, mbps_a,
			(int)best_b, mbps_b);
	}}

	return 0;
}
